"""
"""
import re
import sys
from collections import Counter
from typing import Literal

from pydantic import BaseModel, field_validator

from cognite.neat.rules import exceptions

from ._base import (
    CLASS_ID_REGEX,
    CLASS_ID_REGEX_COMPILED,
    ENTITY_ID_REGEX,
    PROPERTY_ID_REGEX,
    SUFFIX_REGEX,
    Entity,
    EntityTypes,
)

if sys.version_info >= (3, 11):
    from enum import StrEnum
    from typing import Self
else:
    from backports.strenum import StrEnum
    from typing_extensions import Self


class TransformationRuleType(StrEnum):
    rdfpath = "rdfpath"
    rawlookup = "rawlookup"
    sparql = "sparql"


class Lookup(StrEnum):
    table = "table"
    key = "key"
    value = "value"  # type: ignore


# traversal direction
direction_re = r"(?P<direction>(->|<-))"

# steps
step_re = rf"((->|<-){CLASS_ID_REGEX}({PROPERTY_ID_REGEX})?)"
step_re_compiled = re.compile(step_re)
step_class_re_compiled = re.compile(rf"(^{direction_re}{CLASS_ID_REGEX})$")
step_class_and_property_re_compiled = re.compile(rf"(^{direction_re}{CLASS_ID_REGEX}{PROPERTY_ID_REGEX}$)")


_traversal = "traversal"
origin_re = rf"(?P<origin>{ENTITY_ID_REGEX})"

hop_re_compiled = re.compile(rf"^{origin_re}(?P<{_traversal}>{step_re}+)$")

# grabbing specific property for a class, property can be either object, annotation or data property
single_property_re_compiled = re.compile(rf"^{CLASS_ID_REGEX}{PROPERTY_ID_REGEX}$")

# grabbing all properties for a class
all_properties_re_compiled = re.compile(rf"^{CLASS_ID_REGEX}\(\*\)$")

all_traversal_re_compiled = (
    rf"({CLASS_ID_REGEX}\(\*\)|{CLASS_ID_REGEX}{PROPERTY_ID_REGEX}|{origin_re}(?P<{_traversal}>{step_re}+))"
)

table_re_compiled = re.compile(
    rf"^(?P<{Lookup.table}>{SUFFIX_REGEX})\((?P<{Lookup.key}>{SUFFIX_REGEX}),\s*(?P<{Lookup.value}>{SUFFIX_REGEX})\)$"
)


StepDirection = Literal["source", "target", "origin"]
_direction_by_symbol: dict[str, StepDirection] = {"->": "target", "<-": "source"}


class Step(BaseModel):
    class_: Entity
    property: Entity | None = None  # only terminal step has property
    direction: StepDirection

    @classmethod
    def from_string(cls, raw: str, **kwargs) -> Self:
        if result := step_class_and_property_re_compiled.match(raw):
            return cls(
                class_=Entity.from_string(result.group(EntityTypes.class_), type_="class"),
                property=Entity.from_string(result.group(EntityTypes.property_), type_="property"),
                direction=_direction_by_symbol[result.group("direction")],
                **kwargs,
            )
        elif result := step_class_re_compiled.match(raw):
            return cls(
                class_=Entity.from_string(result.group(EntityTypes.class_)),
                direction=_direction_by_symbol[result.group("direction")],
            )  # type: ignore
        msg = f"Invalid step {raw}, expected in one of the following forms:"
        msg += " ->prefix:suffix, <-prefix:suffix, ->prefix:suffix(prefix:suffix) or <-prefix:suffix(prefix:suffix)"
        raise ValueError(msg)


class Traversal(BaseModel):
    class_: Entity


class SingleProperty(Traversal):
    property: Entity

    @classmethod
    def from_string(cls, class_: str, property_: str) -> Self:
        return cls(
            class_=Entity.from_string(class_, type_="class"), property=Entity.from_string(property_, type_="property")
        )


class AllReferences(Traversal):
    @classmethod
    def from_string(cls, class_: str) -> Self:
        return cls(class_=Entity.from_string(class_, type_="class"))


class AllProperties(Traversal):
    @classmethod
    def from_string(cls, class_: str) -> Self:
        return cls(class_=Entity.from_string(class_, type_="class"))


class Origin(BaseModel):
    class_: Entity

    @field_validator("class_", mode="before")
    def process_if_string(cls, value):
        return Entity.from_string(value) if isinstance(value, str) else value


class Hop(Traversal):
    """Multi or single hop traversal through graph"""

    traversal: list[Step]

    @classmethod
    def from_string(cls, class_: str, traversal: str | list[Step]) -> Self:
        return cls(
            class_=Entity.from_string(class_),
            traversal=[Step.from_string(result[0]) for result in step_re_compiled.findall(traversal)]
            if isinstance(traversal, str)
            else traversal,
        )


class TableLookup(BaseModel):
    name: str
    key: str
    value: str


class Rule(BaseModel):
    pass


class Query(BaseModel):
    query: str


class RDFPath(Rule):
    traversal: Traversal | Query


class RawLookup(RDFPath):
    table: TableLookup


class SPARQLQuery(RDFPath):
    traversal: Query


def parse_traversal(raw: str) -> AllReferences | AllProperties | SingleProperty | Hop:
    if result := CLASS_ID_REGEX_COMPILED.match(raw):
        return AllReferences.from_string(class_=result.group(EntityTypes.class_))
    elif result := all_properties_re_compiled.match(raw):
        return AllProperties.from_string(class_=result.group(EntityTypes.class_))
    elif result := single_property_re_compiled.match(raw):
        return SingleProperty.from_string(
            class_=result.group(EntityTypes.class_), property_=result.group(EntityTypes.property_)
        )
    elif result := hop_re_compiled.match(raw):
        return Hop.from_string(class_=result.group("origin"), traversal=result.group(_traversal))
    else:
        raise exceptions.NotValidRDFPath(raw).to_pydantic_custom_error()


def parse_table_lookup(raw: str) -> TableLookup:
    if result := table_re_compiled.match(raw):
        return TableLookup(
            name=result.group(Lookup.table), key=result.group(Lookup.key), value=result.group(Lookup.value)
        )
    raise exceptions.NotValidTableLookUp(raw).to_pydantic_custom_error()


def parse_rule(rule_raw: str, rule_type: TransformationRuleType | None) -> RDFPath:
    match rule_type:
        case TransformationRuleType.rdfpath:
            rule_raw = rule_raw.replace(" ", "")
            return RDFPath(traversal=parse_traversal(rule_raw))
        case TransformationRuleType.rawlookup:
            rule_raw = rule_raw.replace(" ", "")
            if Counter(rule_raw).get("|") != 1:
                raise exceptions.NotValidRAWLookUp(rule_raw).to_pydantic_custom_error()
            traversal, table_lookup = rule_raw.split("|")
            return RawLookup(traversal=parse_traversal(traversal), table=parse_table_lookup(table_lookup))
        case TransformationRuleType.sparql:
            return SPARQLQuery(traversal=Query(query=rule_raw))
        case None:
            raise ValueError("Rule type must be specified")


def is_valid_rule(rule_type: TransformationRuleType, rule_raw: str) -> bool:
    is_valid_rule = {TransformationRuleType.rdfpath: is_rdfpath, TransformationRuleType.rawlookup: is_rawlookup}[
        rule_type
    ]
    return is_valid_rule(rule_raw)


def is_rdfpath(raw: str) -> bool:
    try:
        parse_traversal(raw)
    except ValueError:
        return False
    return True


def is_rawlookup(raw: str) -> bool:
    try:
        parse_rule(raw, TransformationRuleType.rawlookup)
    except ValueError:
        return False
    return True



import json
import warnings
from collections import Counter, defaultdict
from dataclasses import dataclass, field
from typing import ClassVar

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes import DatabaseWrite, DatabaseWriteList, TransformationWrite, TransformationWriteList
from cognite.client.data_classes.data_modeling import ViewApply
from cognite.client.data_classes.transformations.common import Edges, EdgeType, Nodes, ViewInfo

from cognite.neat.rules.issues.dms import (
    ContainerPropertyUsedMultipleTimesError,
    DirectRelationMissingSourceError,
    DMSSchemaError,
    DuplicatedViewInDataModelError,
    MissingContainerError,
    MissingContainerPropertyError,
    MissingEdgeViewError,
    MissingParentViewError,
    MissingSourceViewError,
    MissingSpaceError,
    MissingViewError,
)
from cognite.neat.utils.cdf_loaders import ViewLoader
from cognite.neat.utils.cdf_loaders.data_classes import RawTableWrite, RawTableWriteList


@dataclass
class DMSSchema:
    spaces: dm.SpaceApplyList = field(default_factory=lambda: dm.SpaceApplyList([]))
    data_models: dm.DataModelApplyList = field(default_factory=lambda: dm.DataModelApplyList([]))
    views: dm.ViewApplyList = field(default_factory=lambda: dm.ViewApplyList([]))
    containers: dm.ContainerApplyList = field(default_factory=lambda: dm.ContainerApplyList([]))
    node_types: dm.NodeApplyList = field(default_factory=lambda: dm.NodeApplyList([]))

    @classmethod
    def from_model_id(cls, client: CogniteClient, data_model_id: dm.DataModelIdentifier) -> "DMSSchema":
        data_models = client.data_modeling.data_models.retrieve(data_model_id, inline_views=True)
        if len(data_models) == 0:
            raise ValueError(f"Data model {data_model_id} not found")
        data_model = data_models.latest_version()
        views = dm.ViewList(data_model.views)
        container_ids = views.referenced_containers()
        containers = client.data_modeling.containers.retrieve(list(container_ids))
        cls._append_referenced_containers(client, containers)

        space_read = client.data_modeling.spaces.retrieve(data_model.space)
        if space_read is None:
            raise ValueError(f"Space {data_model.space} not found")
        space = space_read.as_write()
        data_model_write = data_model.as_write()
        data_model_write.views = list(views.as_write())

        # Converting views from read to write format requires to account for parents (implements)
        # as the read format contains all properties from all parents, while the write formate should not contain
        # properties from any parents.
        # The ViewLoader as_write method looks up parents and remove properties from them.
        view_loader = ViewLoader(client)
        view_write = dm.ViewApplyList([view_loader.as_write(view) for view in views])

        return cls(
            spaces=dm.SpaceApplyList([space]),
            data_models=dm.DataModelApplyList([data_model_write]),
            views=view_write,
            containers=containers.as_write(),
        )

    def validate(self) -> list[DMSSchemaError]:
        errors: set[DMSSchemaError] = set()
        defined_spaces = {space.space for space in self.spaces}
        defined_containers = {container.as_id(): container for container in self.containers}
        defined_views = {view.as_id() for view in self.views}

        for container in self.containers:
            if container.space not in defined_spaces:
                errors.add(MissingSpaceError(space=container.space, referred_by=container.as_id()))

        for view in self.views:
            view_id = view.as_id()
            if view.space not in defined_spaces:
                errors.add(MissingSpaceError(space=view.space, referred_by=view_id))

            for parent in view.implements or []:
                if parent not in defined_views:
                    errors.add(MissingParentViewError(view=parent, referred_by=view_id))

            for prop_name, prop in (view.properties or {}).items():
                if isinstance(prop, dm.MappedPropertyApply):
                    ref_container = defined_containers.get(prop.container)
                    if ref_container is None:
                        errors.add(MissingContainerError(container=prop.container, referred_by=view_id))
                    elif prop.container_property_identifier not in ref_container.properties:
                        errors.add(
                            MissingContainerPropertyError(
                                container=prop.container,
                                property=prop.container_property_identifier,
                                referred_by=view_id,
                            )
                        )
                    else:
                        container_property = ref_container.properties[prop.container_property_identifier]

                        if isinstance(container_property.type, dm.DirectRelation) and prop.source is None:
                            errors.add(DirectRelationMissingSourceError(view_id=view_id, property=prop_name))

                if isinstance(prop, dm.EdgeConnectionApply) and prop.source not in defined_views:
                    errors.add(MissingSourceViewError(view=prop.source, property=prop_name, referred_by=view_id))

                if (
                    isinstance(prop, dm.EdgeConnectionApply)
                    and prop.edge_source is not None
                    and prop.edge_source not in defined_views
                ):
                    errors.add(MissingEdgeViewError(view=prop.edge_source, property=prop_name, referred_by=view_id))

            property_count = Counter(
                (prop.container, prop.container_property_identifier)
                for prop in (view.properties or {}).values()
                if isinstance(prop, dm.MappedPropertyApply)
            )
            for (container_id, container_property_identifier), count in property_count.items():
                if count > 1:
                    view_properties = [
                        prop_name
                        for prop_name, prop in (view.properties or {}).items()
                        if isinstance(prop, dm.MappedPropertyApply)
                        and (prop.container, prop.container_property_identifier)
                        == (container_id, container_property_identifier)
                    ]
                    errors.add(
                        ContainerPropertyUsedMultipleTimesError(
                            container=container_id,
                            property=container_property_identifier,
                            referred_by=frozenset({(view_id, prop_name) for prop_name in view_properties}),
                        )
                    )

        for model in self.data_models:
            if model.space not in defined_spaces:
                errors.add(MissingSpaceError(space=model.space, referred_by=model.as_id()))

            view_counts: dict[dm.ViewId, int] = defaultdict(int)
            for view_id_or_class in model.views or []:
                view_id = view_id_or_class if isinstance(view_id_or_class, dm.ViewId) else view_id_or_class.as_id()
                if view_id not in defined_views:
                    errors.add(MissingViewError(referred_by=model.as_id(), view=view_id))
                view_counts[view_id] += 1

            for view_id, count in view_counts.items():
                if count > 1:
                    errors.add(DuplicatedViewInDataModelError(referred_by=model.as_id(), view=view_id))

        return list(errors)

    @classmethod
    def _append_referenced_containers(cls, client: CogniteClient, containers: dm.ContainerList) -> None:
        """Containers can reference each other through the 'requires' constraint.

        This method retrieves all containers that are referenced by other containers through the 'requires' constraint,
        including their parents.

        """
        for _ in range(10):  # Limiting the number of iterations to avoid infinite loops
            referenced_containers = {
                const.require
                for container in containers
                for const in (container.constraints or {}).values()
                if isinstance(const, dm.RequiresConstraint)
            }
            missing_containers = referenced_containers - set(containers.as_ids())
            if not missing_containers:
                break
            found_containers = client.data_modeling.containers.retrieve(list(missing_containers))
            containers.extend(found_containers)
            if len(found_containers) != len(missing_containers):
                break
        else:
            warnings.warn(
                "The maximum number of iterations was reached while resolving referenced containers."
                "There might be referenced containers that are not included in the list of containers.",
                RuntimeWarning,
                stacklevel=2,
            )
        return None


@dataclass
class PipelineSchema(DMSSchema):
    transformations: TransformationWriteList = field(default_factory=lambda: TransformationWriteList([]))
    databases: DatabaseWriteList = field(default_factory=lambda: DatabaseWriteList([]))
    raw_tables: RawTableWriteList = field(default_factory=lambda: RawTableWriteList([]))

    _SQL_TYPE_BY_PROPERTY_TYPE: ClassVar[dict[type[dm.PropertyType], str]] = {
        dm.Text: "STRING",
        dm.Int32: "INTEGER",
        dm.Int64: "INTEGER",
        dm.Float32: "FLOAT",
        dm.Float64: "DOUBLE",
        dm.Date: "DATE",
        dm.Timestamp: "TIMESTAMP",
        dm.TimeSeriesReference: "STRING",
        dm.FileReference: "STRING",
        dm.SequenceReference: "STRING",
        dm.Boolean: "BOOLEAN",
        dm.Json: "STRING",
    }

    @classmethod
    def from_dms(cls, schema: DMSSchema, instance_space: str | None = None) -> "PipelineSchema":
        if not schema.data_models:
            raise ValueError("PipelineSchema must contain at least one data model")
        first_data_model = schema.data_models[0]
        # The database name is limited to 32 characters
        database_name = first_data_model.external_id[:32]
        instance_space = instance_space or first_data_model.space
        database = DatabaseWrite(name=database_name)
        parent_views = {parent for view in schema.views for parent in view.implements or []}
        container_by_id = {container.as_id(): container for container in schema.containers}

        transformations = TransformationWriteList([])
        raw_tables = RawTableWriteList([])
        for view in schema.views:
            if view.as_id() in parent_views:
                # Skipping parents as they do not have their own data
                continue
            mapped_properties = {
                prop_name: prop
                for prop_name, prop in (view.properties or {}).items()
                if isinstance(prop, dm.MappedPropertyApply)
            }
            if mapped_properties:
                view_table = RawTableWrite(name=f"{view.external_id}Properties", database=database_name)
                raw_tables.append(view_table)
                transformation = cls._create_property_transformation(
                    mapped_properties, view, view_table, container_by_id, instance_space
                )
                transformations.append(transformation)
            connection_properties = {
                prop_name: prop
                for prop_name, prop in (view.properties or {}).items()
                if isinstance(prop, dm.EdgeConnectionApply)
            }
            for prop_name, connection_property in connection_properties.items():
                view_table = RawTableWrite(name=f"{view.external_id}.{prop_name}Edge", database=database_name)
                raw_tables.append(view_table)
                transformation = cls._create_edge_transformation(connection_property, view, view_table, instance_space)
                transformations.append(transformation)

        return cls(
            spaces=schema.spaces,
            data_models=schema.data_models,
            views=schema.views,
            containers=schema.containers,
            transformations=transformations,
            databases=DatabaseWriteList([database]),
            raw_tables=raw_tables,
        )

    @classmethod
    def _create_property_transformation(
        cls,
        properties: dict[str, dm.MappedPropertyApply],
        view: ViewApply,
        table: RawTableWrite,
        container_by_id: dict[dm.ContainerId, dm.ContainerApply],
        instance_space: str,
    ) -> TransformationWrite:
        mapping_mode = {
            "version": 1,
            "sourceType": "raw",
            # 'mappings' is set here and overwritten further down to ensure the correct order
            "mappings": [],
            "sourceLevel1": table.database,
            "sourceLevel2": table.name,
        }
        mappings = [
            {"from": "externalId", "to": "externalId", "asType": "STRING"},
        ]
        select_rows = ["cast(`externalId` as STRING) as externalId"]
        for prop_name, prop in properties.items():
            container = container_by_id.get(prop.container)
            if container is not None:
                sql_type = (
                    cls._SQL_TYPE_BY_PROPERTY_TYPE.get(
                        type(container.properties[prop.container_property_identifier].type)
                    )
                    or "STRING"
                )
            else:
                sql_type = "STRING"
            select_rows.append(f"cast(`{prop_name}` as {sql_type}) as {prop_name}")
            mappings.append({"from": prop_name, "to": prop_name, "asType": sql_type})
        mapping_mode["mappings"] = mappings
        select = ",\n  ".join(select_rows)

        return TransformationWrite(
            external_id=f"{table.name}Transformation",
            name=f"{table.name}Transformation",
            ignore_null_fields=True,
            destination=Nodes(
                view=ViewInfo(view.space, view.external_id, view.version),
                instance_space=instance_space,
            ),
            conflict_mode="upsert",
            query=f"""/* MAPPING_MODE_ENABLED: true */
/* {json.dumps(mapping_mode)} */
select
  {select}
from
  `{table.database}`.`{table.name}`;
""",
        )

    @classmethod
    def _create_edge_transformation(
        cls, property_: dm.EdgeConnectionApply, view: ViewApply, table: RawTableWrite, instance_space: str
    ) -> TransformationWrite:
        start, end = view.external_id, property_.source.external_id
        if property_.direction == "inwards":
            start, end = end, start
        mapping_mode = {
            "version": 1,
            "sourceType": "raw",
            "mappings": [
                {"from": "externalId", "to": "externalId", "asType": "STRING"},
                {"from": start, "to": "startNode", "asType": "STRUCT<`space`:STRING, `externalId`:STRING>"},
                {"from": end, "to": "endNode", "asType": "STRUCT<`space`:STRING, `externalId`:STRING>"},
            ],
            "sourceLevel1": table.database,
            "sourceLevel2": table.name,
        }
        select_rows = [
            "cast(`externalId` as STRING) as externalId",
            f"node_reference('{instance_space}', `{start}`) as startNode",
            f"node_reference('{instance_space}', `{end}`) as endNode",
        ]
        select = ",\n  ".join(select_rows)

        return TransformationWrite(
            external_id=f"{table.name}Transformation",
            name=f"{table.name}Transformation",
            ignore_null_fields=True,
            destination=Edges(
                instance_space=instance_space,
                edge_type=EdgeType(space=property_.type.space, external_id=property_.type.external_id),
            ),
            conflict_mode="upsert",
            query=f"""/* MAPPING_MODE_ENABLED: true */
/* {json.dumps(mapping_mode)} */
select
  {select}
from
  `{table.database}`.`{table.name}`;
""",
        )



"""To be renamed to non-private module once the migration is completed."""

import itertools
import warnings
from collections import defaultdict

import pandas as pd

from cognite.neat.rules.models._rules._types import ClassEntity, EntityTypes, ParentClassEntity
from cognite.neat.rules.models._rules.information_rules import InformationClass, InformationProperty, InformationRules
from cognite.neat.rules.models.rdfpath import TransformationRuleType
from cognite.neat.utils.utils import get_inheritance_path


def get_class_parent_pairs(rule: InformationRules) -> dict[ClassEntity, list[ParentClassEntity]]:
    """This only returns class - parent pairs only if parent is in the same data model"""
    class_subclass_pairs: dict[ClassEntity, list[ParentClassEntity]] = {}
    for definition in rule.classes:
        class_subclass_pairs[definition.class_] = []

        if definition.parent is None:
            continue

        for parent in definition.parent:
            if parent.prefix == definition.class_.prefix:
                class_subclass_pairs[definition.class_].append(parent)
            else:
                warnings.warn(
                    f"Parent class {parent} of class {definition} is not in the same namespace, skipping !",
                    stacklevel=2,
                )

    return class_subclass_pairs


def get_classes_with_properties(
    rules: InformationRules, consider_inheritance: bool = False
) -> dict[ClassEntity, list[InformationProperty]]:
    """Returns classes that have been defined in the data model.

    Args:
        rules: Instance of InformationRules holding the data model
        consider_inheritance: Whether to consider inheritance or not. Defaults False

    Returns:
        Dictionary of classes with a list of properties defined for them

    !!! note "consider_inheritance"
        If consider_inheritance is True, properties from parent classes will also be considered.
        This means if a class has a parent class, and the parent class has properties defined for it,
        while we do not have any properties defined for the child class, we will still consider the
        properties from the parent class. If consider_inheritance is False, we will only consider
        properties defined for the child class, thus if no properties are defined for the child class,
        it will not be included in the returned dictionary.
    """

    class_property_pairs: dict[ClassEntity, list[InformationProperty]] = {}

    for property_ in rules.properties:
        if property_.class_ in class_property_pairs:
            class_property_pairs[property_.class_] += [property_]
        else:
            class_property_pairs[property_.class_] = [property_]

    if consider_inheritance:
        class_parent_pairs = get_class_parent_pairs(rules)
        for class_ in class_parent_pairs:
            _add_inherited_properties(class_, class_property_pairs, class_parent_pairs)

    return class_property_pairs


def _add_inherited_properties(class_, class_property_pairs, class_parent_pairs):
    inheritance_path = get_inheritance_path(class_, class_parent_pairs)
    for parent in inheritance_path:
        # ParentClassEntity -> ClassEntity to match the type of class_property_pairs
        if parent.as_class_entity() in class_property_pairs:
            for property_ in class_property_pairs[parent.as_class_entity()]:
                property_ = property_.model_copy()
                # this corresponds to importing properties from parent class
                property_.class_ = class_

                if class_ in class_property_pairs:
                    class_property_pairs[class_].append(property_)
                else:
                    class_property_pairs[class_] = [property_]


def to_class_property_pairs(
    rules: InformationRules, only_rdfpath: bool = False, consider_inheritance: bool = False
) -> dict[ClassEntity, dict[str, InformationProperty]]:
    """Returns a dictionary of classes with a dictionary of properties associated with them.

    Args:
        rules : Instance of InformationRules holding the data model
        only_rdfpath : To consider only properties which have rule `rdfpath` set. Defaults False
        consider_inheritance: Whether to consider inheritance or not. Defaults False

    Returns:
        Dictionary of classes with a dictionary of properties associated with them.

    !!! note "difference to get_classes_with_properties"
        This method returns a dictionary of classes with a dictionary of properties associated with them.
        While get_classes_with_properties returns a dictionary of classes with a list of properties defined for them,
        here we filter the properties based on the `only_rdfpath` parameter and only consider
        the first definition of a property if it is defined more than once.

    !!! note "only_rdfpath"
        If only_rdfpath is True, only properties with RuleType.rdfpath will be returned as
        a part of the dictionary of properties related to a class. Otherwise, all properties
        will be returned.

    !!! note "consider_inheritance"
        If consider_inheritance is True, properties from parent classes will also be considered.
        This means if a class has a parent class, and the parent class has properties defined for it,
        while we do not have any properties defined for the child class, we will still consider the
        properties from the parent class. If consider_inheritance is False, we will only consider
        properties defined for the child class, thus if no properties are defined for the child class,
        it will not be included in the returned dictionary.
    """
    # TODO: https://cognitedata.atlassian.net/jira/software/projects/NEAT/boards/893?selectedIssue=NEAT-78

    class_property_pairs = {}

    for class_, properties in get_classes_with_properties(rules, consider_inheritance).items():
        processed_properties = {}
        for property_ in properties:
            if property_.property_ in processed_properties:
                # TODO: use appropriate Warning class from _exceptions.py
                # if missing make one !
                warnings.warn(
                    f"Property {property_.property_} for {class_} has been defined more than once!"
                    " Only the first definition will be considered, skipping the rest..",
                    stacklevel=2,
                )
                continue

            if (only_rdfpath and property_.rule_type == TransformationRuleType.rdfpath) or not only_rdfpath:
                processed_properties[property_.property_] = property_
        class_property_pairs[class_] = processed_properties

    return class_property_pairs


def get_class_linkage(rules: InformationRules, consider_inheritance: bool = False) -> pd.DataFrame:
    """Returns a dataframe with the class linkage of the data model.

    Args:
        rules: Instance of InformationRules holding the data model
        consider_inheritance: Whether to consider inheritance or not. Defaults False

    Returns:
        Dataframe with the class linkage of the data model
    """

    class_linkage = pd.DataFrame(columns=["source_class", "target_class", "connecting_property", "max_occurrence"])

    class_property_pairs = get_classes_with_properties(rules, consider_inheritance)
    properties = list(itertools.chain.from_iterable(class_property_pairs.values()))

    for property_ in properties:
        if property_.type_ == EntityTypes.object_property:
            new_row = pd.Series(
                {
                    "source_class": property_.class_,
                    "connecting_property": property_.property_,
                    "target_class": property_.value_type,
                    "max_occurrence": property_.max_count,
                }
            )
            class_linkage = pd.concat([class_linkage, new_row.to_frame().T], ignore_index=True)

    class_linkage.drop_duplicates(inplace=True)
    class_linkage = class_linkage[["source_class", "connecting_property", "target_class", "max_occurrence"]]

    return class_linkage


def get_connected_classes(rules: InformationRules, consider_inheritance: bool = False) -> set[ClassEntity]:
    """Return a set of classes that are connected to other classes.

    Args:
        rules: Instance of InformationRules holding the data model
        consider_inheritance: Whether to consider inheritance or not. Defaults False

    Returns:
        Set of classes that are connected to other classes
    """
    class_linkage = get_class_linkage(rules, consider_inheritance)
    return set(class_linkage.source_class.values).union(set(class_linkage.target_class.values))


def get_defined_classes(rules: InformationRules, consider_inheritance: bool = False) -> set[ClassEntity]:
    """Returns classes that have properties defined for them in the data model.

    Args:
        rules: Instance of InformationRules holding the data model
        consider_inheritance: Whether to consider inheritance or not. Defaults False

    Returns:
        Set of classes that have been defined in the data model
    """
    class_property_pairs = get_classes_with_properties(rules, consider_inheritance)
    properties = list(itertools.chain.from_iterable(class_property_pairs.values()))

    return {property.class_ for property in properties}


def get_disconnected_classes(rules: InformationRules, consider_inheritance: bool = False) -> set[ClassEntity]:
    """Return a set of classes that are disconnected (i.e. isolated) from other classes.

    Args:
        rules: Instance of InformationRules holding the data model
        consider_inheritance: Whether to consider inheritance or not. Defaults False

    Returns:
        Set of classes that are disconnected from other classes
    """
    return get_defined_classes(rules, consider_inheritance) - get_connected_classes(rules, consider_inheritance)


def get_symmetric_pairs(
    rules: InformationRules, consider_inheritance: bool = False
) -> set[tuple[ClassEntity, ClassEntity]]:
    """Returns a set of pairs of symmetrically linked classes.

    Args:
        rules: Instance of InformationRules holding the data model
        consider_inheritance: Whether to consider inheritance or not. Defaults False

    Returns:
        Set of pairs of symmetrically linked classes
    """

    # TODO: Find better name for this method
    sym_pairs: set[tuple[ClassEntity, ClassEntity]] = set()

    class_linkage = get_class_linkage(rules, consider_inheritance)
    if class_linkage.empty:
        return sym_pairs

    for _, row in class_linkage.iterrows():
        source = row.source_class
        target = row.target_class
        target_targets = class_linkage[class_linkage.source_class == target].target_class.values
        if source in target_targets and (source, target) not in sym_pairs:
            sym_pairs.add((source, target))
    return sym_pairs


def to_property_dict(rules: InformationRules) -> dict[str, list[InformationProperty]]:
    """This is used to capture all definitions of a property in the data model."""
    property_dict: dict[str, list[InformationProperty]] = defaultdict(list)
    for definition in rules.properties:
        property_dict[definition.property_].append(definition)
    return property_dict


def to_class_dict(rules: InformationRules) -> dict[str, InformationClass]:
    """This is to simplify access to classes through dict."""
    class_dict: dict[str, InformationClass] = {}
    for definition in rules.classes:
        class_dict[definition.class_.suffix] = definition
    return class_dict



from typing import Any, ClassVar

from pydantic import Field, model_serializer
from pydantic_core.core_schema import SerializationInfo

from ._types import ParentClassType, PropertyType, SemanticValueType, StrOrListType
from .base import (
    BaseMetadata,
    RoleTypes,
    RuleModel,
    SheetEntity,
    SheetList,
)


class DomainMetadata(BaseMetadata):
    role: ClassVar[RoleTypes] = RoleTypes.domain_expert
    creator: StrOrListType


class DomainProperty(SheetEntity):
    property_: PropertyType = Field(alias="Property")
    value_type: SemanticValueType = Field(alias="Value Type")
    min_count: int | None = Field(alias="Min Count", default=None)
    max_count: int | float | None = Field(alias="Max Count", default=None)


class DomainClass(SheetEntity):
    description: str | None = Field(None, alias="Description")
    parent: ParentClassType = Field(alias="Parent Class")


class DomainRules(RuleModel):
    metadata: DomainMetadata = Field(alias="Metadata")
    properties: SheetList[DomainProperty] = Field(alias="Properties")
    classes: SheetList[DomainClass] | None = Field(None, alias="Classes")

    @model_serializer(mode="plain", when_used="always")
    def domain_rules_serializer(self, info: SerializationInfo) -> dict[str, Any]:
        kwargs = vars(info)
        output: dict[str, Any] = {
            "Metadata" if info.by_alias else "metadata": self.metadata.model_dump(**kwargs),
            "Properties" if info.by_alias else "properties": [prop.model_dump(**kwargs) for prop in self.properties],
        }
        if self.classes or not info.exclude_none:
            output["Classes" if info.by_alias else "classes"] = [
                cls.model_dump(**kwargs) for cls in self.classes or []
            ] or None
        return output



"""This module provides set of methods that perform conversion of TransformationRules
to TransformationRules for purpose of for example:

- subsetting the data model to only include desired classes and their properties
- converting classes/properties ids/names to DMS compliant format
"""

import logging
import warnings
from typing import Any

from pydantic import ValidationError

from cognite.neat.rules._analysis import (
    get_class_parent_pairs,
    get_classes_with_properties,
    get_defined_classes,
    to_class_dict,
)
from cognite.neat.rules.models._rules import InformationRules
from cognite.neat.rules.models._rules._types import ClassEntity
from cognite.neat.rules.models._rules.base import SchemaCompleteness
from cognite.neat.utils.utils import get_inheritance_path


def subset_rules(rules: InformationRules, desired_classes: set[ClassEntity]) -> InformationRules:
    """
    Subset rules to only include desired classes and their properties.

    Args:
        rules: Instance of InformationRules to subset
        desired_classes: Desired classes to include in the reduced data model

    Returns:
        Instance of InformationRules

    !!! note "Inheritance"
        If desired classes contain a class that is a subclass of another class(es), the parent class(es)
        will be included in the reduced data model as well even though the parent class(es) are
        not in the desired classes set. This is to ensure that the reduced data model is
        consistent and complete.

    !!! note "Partial Reduction"
        This method does not perform checks if classes that are value types of desired classes
        properties are part of desired classes. If a class is not part of desired classes, but it
        is a value type of a property of a class that is part of desired classes, derived reduced
        rules will be marked as partial.

    !!! note "Validation"
        This method will attempt to validate the reduced rules with custom validations.
        If it fails, it will return a partial rules with a warning message, validated
        only with base Pydantic validators.
    """

    if not rules.metadata.schema_ is not SchemaCompleteness.complete:
        raise ValueError("Rules are not complete cannot perform reduction!")
    class_as_dict = to_class_dict(rules)
    class_parents_pairs = get_class_parent_pairs(rules)
    defined_classes = get_defined_classes(rules, True)

    possible_classes = defined_classes.intersection(desired_classes)
    impossible_classes = desired_classes - possible_classes

    # need to add all the parent classes of the desired classes to the possible classes
    parents: set[ClassEntity] = set()
    for class_ in possible_classes:
        parents = parents.union(
            {parent.as_class_entity() for parent in get_inheritance_path(class_, class_parents_pairs)}
        )
    possible_classes = possible_classes.union(parents)

    if not possible_classes:
        logging.error("None of the desired classes are defined in the data model!")
        raise ValueError("None of the desired classes are defined in the data model!")

    if impossible_classes:
        logging.warning(f"Could not find the following classes defined in the data model: {impossible_classes}")
        warnings.warn(
            f"Could not find the following classes defined in the data model: {impossible_classes}", stacklevel=2
        )

    reduced_data_model: dict[str, Any] = {
        "metadata": rules.metadata.model_copy(),
        "prefixes": (rules.prefixes or {}).copy(),
        "classes": [],
        "properties": [],
    }

    logging.info(f"Reducing data model to only include the following classes: {possible_classes}")
    for class_ in possible_classes:
        reduced_data_model["classes"].append(class_as_dict[class_.suffix])

    class_property_pairs = get_classes_with_properties(rules, False)

    for class_, properties in class_property_pairs.items():
        if class_ in possible_classes:
            reduced_data_model["properties"].extend(properties)

    try:
        return InformationRules(**reduced_data_model)
    except ValidationError as e:
        warnings.warn(f"Reduced data model is not complete: {e}", stacklevel=2)
        reduced_data_model["metadata"].schema_ = SchemaCompleteness.partial
        return InformationRules.model_construct(**reduced_data_model)



"""This module performs importing of graph to TransformationRules pydantic class.
In more details, it traverses the graph and abstracts class and properties, basically
generating a list of rules based on which nodes that form the graph are made.
"""
from collections import defaultdict
from pathlib import Path
from typing import Literal, cast, overload

import pandas as pd

import cognite.neat.rules.issues.spreadsheet_file
from cognite.neat.rules import issues
from cognite.neat.rules.issues import IssueList
from cognite.neat.rules.models._rules import RULES_PER_ROLE, DMSRules, DomainRules, InformationRules
from cognite.neat.rules.models._rules.base import RoleTypes
from cognite.neat.utils.auxiliary import local_import
from cognite.neat.utils.spreadsheet import SpreadsheetRead, read_spreadsheet

from ._base import BaseImporter, Rules, _handle_issues


class ExcelImporter(BaseImporter):
    def __init__(self, filepath: Path):
        self.filepath = filepath

    @overload
    def to_rules(self, errors: Literal["raise"], role: RoleTypes | None = None) -> Rules:
        ...

    @overload
    def to_rules(
        self, errors: Literal["continue"] = "continue", role: RoleTypes | None = None
    ) -> tuple[Rules | None, IssueList]:
        ...

    def to_rules(
        self, errors: Literal["raise", "continue"] = "continue", role: RoleTypes | None = None
    ) -> tuple[Rules | None, IssueList] | Rules:
        issue_list = IssueList(title=f"'{self.filepath.name}'")
        if not self.filepath.exists():
            issue_list.append(cognite.neat.rules.issues.spreadsheet_file.SpreadsheetNotFoundError(self.filepath))
            if errors == "raise":
                raise issue_list.as_errors() from None
            return None, issue_list

        with pd.ExcelFile(self.filepath) as excel_file:
            try:
                metadata = dict(pd.read_excel(excel_file, "Metadata", header=None).values)
            except ValueError:
                issue_list.append(
                    cognite.neat.rules.issues.spreadsheet_file.MetadataSheetMissingOrFailedError(self.filepath)
                )
                if errors == "raise":
                    raise issue_list.as_errors() from None
                return None, issue_list

            role_input = RoleTypes(metadata.get("role", RoleTypes.domain_expert))
            role_enum = RoleTypes(role_input)
            rules_model = RULES_PER_ROLE[role_enum]
            sheet_names = {str(name) for name in excel_file.sheet_names}
            expected_sheet_names = rules_model.mandatory_fields(use_alias=True)

            if missing_sheets := expected_sheet_names.difference(sheet_names):
                issue_list.append(
                    cognite.neat.rules.issues.spreadsheet_file.SheetMissingError(self.filepath, list(missing_sheets))
                )
                if errors == "raise":
                    raise issue_list.as_errors()
                return None, issue_list

            sheets: dict[str, dict | list] = {"Metadata": metadata}
            read_info_by_sheet: dict[str, SpreadsheetRead] = defaultdict(SpreadsheetRead)
            for sheet_name, headers in [
                ("Properties", "Class"),
                ("Classes", "Class"),
                ("Containers", "Container"),
                ("Views", "View"),
            ]:
                if sheet_name in excel_file.sheet_names:
                    try:
                        sheets[sheet_name], read_info_by_sheet[sheet_name] = read_spreadsheet(
                            excel_file, sheet_name, return_read_info=True, expected_headers=[headers]
                        )
                    except Exception as e:
                        issue_list.append(
                            cognite.neat.rules.issues.spreadsheet_file.ReadSpreadsheetsError(self.filepath, str(e))
                        )
                        continue
            if issue_list:
                if errors == "raise":
                    raise issue_list.as_errors()
                return None, issue_list

        rules_cls = {
            RoleTypes.domain_expert: DomainRules,
            RoleTypes.information_architect: InformationRules,
            RoleTypes.dms_architect: DMSRules,
        }.get(role_enum)
        if rules_cls is None:
            issue_list.append(cognite.neat.rules.issues.spreadsheet_file.InvalidRoleError(str(role_input)))
            if errors == "raise":
                raise issue_list.as_errors()
            return None, issue_list

        with _handle_issues(
            issue_list,
            error_cls=issues.spreadsheet.InvalidSheetError,
            error_args={"read_info_by_sheet": read_info_by_sheet},
        ) as future:
            rules = rules_cls.model_validate(sheets)  # type: ignore[attr-defined]
        if future.result == "failure":
            if errors == "continue":
                return None, issue_list
            else:
                raise issue_list.as_errors()

        return self._to_output(rules, issue_list, errors=errors, role=role)


class GoogleSheetImporter(BaseImporter):
    def __init__(self, sheet_id: str, skiprows: int = 1):
        self.sheet_id = sheet_id
        self.skiprows = skiprows

    @overload
    def to_rules(self, errors: Literal["raise"], role: RoleTypes | None = None) -> Rules:
        ...

    @overload
    def to_rules(
        self, errors: Literal["continue"] = "continue", role: RoleTypes | None = None
    ) -> tuple[Rules | None, IssueList]:
        ...

    def to_rules(
        self, errors: Literal["raise", "continue"] = "continue", role: RoleTypes | None = None
    ) -> tuple[Rules | None, IssueList] | Rules:
        local_import("gspread", "google")
        import gspread  # type: ignore[import]

        role = role or RoleTypes.domain_expert
        rules_model = cast(DomainRules | InformationRules | DMSRules, RULES_PER_ROLE[role])

        client_google = gspread.service_account()
        google_sheet = client_google.open_by_key(self.sheet_id)
        sheets = {worksheet.title: pd.DataFrame(worksheet.get_all_records()) for worksheet in google_sheet.worksheets()}
        sheet_names = {str(name).lower() for name in sheets.keys()}

        if missing_sheets := rules_model.mandatory_fields().difference(sheet_names):
            raise ValueError(f"Missing mandatory sheets: {missing_sheets}")

        if role == RoleTypes.domain_expert:
            return rules_model.model_validate(sheets)
        elif role == RoleTypes.information_architect:
            return rules_model.model_validate(sheets)
        elif role == RoleTypes.dms_architect:
            return rules_model.model_validate(sheets)
        else:
            raise ValueError(f"Role {role} is not valid.")
